/*
 * train_corr_coeff_vectors.cxx
 *
 *  Created on: Dec 30, 2010
 *      Author: Ignacio Laguna
 *     Contact: ilaguna@purdue.edu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>
#include <math.h>

#include <string>
#include <iostream>
#include <vector>

using namespace std;


/*
 * Number of windows
 */
unsigned int num_windows = 0;

/*
 * Number of metrics
 */
unsigned int num_metrics = 0;

/*
 * Number of data points per window
 */
unsigned int num_points = 0;

/*
 * Error-checking malloc
 */
void * xmalloc (size_t size)
{
  register void *value = malloc(size);
  if (value == 0) {
    cerr << "virtual memory exhausted" << endl;
    exit(-1);
  }
  return value;
}

inline
void handle_error(const char *message)
{
	cerr << string("[detection_engine]: ") + message << endl;
	cerr.flush();
	exit(EXIT_FAILURE);
}

void Tokenize(const string& str, vector<string>& tokens, const string& delimiters = " ")
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

/*
 * Factorial function
 */
inline
unsigned int fact(unsigned int n)
{
	unsigned int ret = 0;

	unsigned int i=0,fact=1;
	if(n <= 1) {
		ret = 1;
	} else {
		for(i=1; i <= n; i++) {
			fact = fact*i;
		}
		ret = fact;
	}

	//printf("Returning: %d\n", (int)ret);
	return ret;
}

/*
 * (n choose k) function
 * It is optimized to avoid integer overflows due to large numbers
 */
inline
unsigned int n_choose_k(unsigned int n, unsigned int k)
{
	unsigned int ret=1, num=1, den=1;

	for (unsigned int i=n; i >= (n-(k-1)); i--)
		num = num * i;

	den = fact(k); // I do factorial of this because it's a small number
	ret = num / den;

	//printf("Num=%d, Den=%d, ret=%d\n", num, den, ret);
	return ret;
}

/*
 * Set initial parameters using a Perl script
 */
void set_params(const string &path)
{
	size_t n = 100;
	char *lineptr = (char *)malloc(n + 1);
	FILE *fd;
	string command = "./get_parameters.pl " + path;

	/*
	 * Open pipe
	 */
	if (!(fd = popen(command.c_str(), "r"))) {
		cerr << "Could not open pipe for " + command << endl;
		exit(-1);
	}

	int count = 0;
	vector<string> tokens;
	while (getline(&lineptr, &n, fd) != -1) {
		lineptr[strlen(lineptr)-1] = '\0'; // eliminate \n character
		Tokenize(string(lineptr), tokens, "|");
		count++;
	}
	pclose(fd);
	free(lineptr);

	if (count == 1) {
		num_windows = static_cast<unsigned int>(atoi(tokens[0].c_str()));
		num_metrics = static_cast<unsigned int>(atoi(tokens[1].c_str()));
		num_points = static_cast<unsigned int>(atoi(tokens[2].c_str()));
	} else {
		cerr << "Error in set_params(): didn't get one line from the Perl script: " << count << endl;
		exit(-1);
	}

	return;
}

/*
 * Initialize windows data structure.
 */
double *** init_windows(void)
{
	double ***wins;
	wins = (double ***) xmalloc(sizeof(double **) * num_windows);
	unsigned int i,j;

	for (i=0; i < num_windows; i++) {
		wins[i] = (double **)xmalloc(sizeof(double *) * num_metrics);
		for (j=0; j < num_metrics; j++) {
			wins[i][j] = (double *)xmalloc(sizeof(double) * num_points);
		}
	}

	return wins;
}

/*
 * Initialize original-data structure.
 */
double *** init_data(void)
{
	unsigned int num_workloads = 3;
	unsigned int num_values = 26640;
	double ***data;
	data = (double ***) xmalloc(sizeof(double **) * num_workloads);
	unsigned int i,j;

	for (i=0; i < num_workloads; i++) {
		data[i] = (double **)xmalloc(sizeof(double *) * num_metrics);
		for (j=0; j < num_metrics; j++) {
			data[i][j] = (double *)xmalloc(sizeof(double) * num_values);
		}
	}

	return data;
}

/*
 * Load windows data files into memory.
 * The files are taken from those that are generated
 * when training ARIMA models with R.
 */
void load_windows(double ***wins, const string &path)
{
	unsigned int i, j, k;
	//cout << "Path from function: " << path << " num_windows: " << num_windows << endl;

	/*
	 * Char buffer for lines of the files
	 */
	size_t n = 50;
	char *lineptr;

	double data;
	FILE *fd;
	string filename;
	char w[10], m[10];

#ifdef WITH_OPENMP
#pragma omp parallel for shared(wins,num_windows,num_metrics,num_points,path) private(i,j,k,data,fd,filename,w,m,lineptr,n)
#endif
	for (i=0; i < num_windows; i++) {
		/* Get window */
		w[0] = '\0';
		sprintf(w, "%d", (int)(i+1));

		for (j=0; j < num_metrics; j++) {
			/* Get metric */
			m[0] = '\0';
			sprintf(m, "%d", (int)(j+1));
			filename = path + "/windows/win_" + string(m) + "_" + string(w);

			fd = fopen(filename.c_str(), "r");
			if (fd == NULL) {
				cerr << "Could not open file " + filename << endl;
			}

			lineptr = (char *)malloc(n + 1);
			k=0;
			while (getline(&lineptr, &n, fd) != -1) {
				//cout << filename << endl;
				data = atof(lineptr);
				wins[i][j][k] = data;
				k++;
			}
			free(lineptr);
			fclose(fd);
		}
	}

	return;
}

/*
 * Delete the last character of a string.
 */
void del_last_char(char *line)
{
	size_t n = strlen(line);
	line[n-1] = '\0';
	return;
}

/**
 * Input structure:
 * data[i][j][k]:
 * i: workload
 * j: metric
 * k: value
 */
void load_original_data(double ***data)
{
	size_t n = 5000;
	char *lineptr;
	FILE *fd;
	//string filename = string("./normal_behavior_original");
	string filename = string("./normal-behavior-newcapacity-20100125");

	fd = fopen(filename.c_str(), "r");
	if (fd == NULL) {
		cerr << "Could not open file " + filename << endl;
	}

	lineptr = (char *)xmalloc(n);

	unsigned int i=0,j,k;
	while (getline(&lineptr, &n, fd) != -1) {
		del_last_char(lineptr);
		string line(lineptr);
		vector<string> tokens;
		Tokenize(line, tokens, string(","));

		unsigned int num_values = 26640;
		for (j=0; j < num_metrics; j++) {
			for (k=0; k < num_values; k++) {
				data[i][j][k] = atof(tokens[(j * num_values) + k].c_str());
			}
		}

		i++;
	}
	free(lineptr);

	return;
}

/*
 * Determine what metrics are constants.
 * Print results into a file.
 */
void check_const_metrics(unsigned int *const_metrics, double *last_values, double ***wins, const string &path)
{
	unsigned int i, j, k;

	for (i=0; i < num_metrics; i++) {
		int is_const = 1;
		double last_v = 0; // last value

		for (j=0; j < num_windows; j++) {
			double x = wins[j][i][0]; // the previous point in the window

			for (k=1; k < num_points; k++) {
				double y = wins[j][i][k]; // the next point in the window
				last_v = y;

				if (x != y) {
					is_const = 0;
					break;
				}

				x = y;
			}
			if (!is_const) break;
		}

		if (is_const) const_metrics[i] = 1;
		else const_metrics[i] = 0;

		last_values[i] = last_v;
	}

	/*
	 * Write the data to a file.
	 * The value in the file is the last one observed. If the metric
	 * is non-constant, it will be the last value before determining
	 * the metrics is constant (i.e., before the loop break).
	 */
	FILE *fd;
	string filename = path + string("/constant_metrics.dat");
	fd = fopen(filename.c_str(), "w");
	if (fd == NULL) {
		cerr << "Could not open file " + filename << endl;
	}

	fputs("#metric,[0|1],value\n",fd);
	for (i=0; i < num_metrics; i++) {
		char buffer[50];
		buffer[0] = '\0';
		sprintf(buffer, "%d,%d,%e\n",(i+1), const_metrics[i], last_values[i]);
		fputs(buffer,fd);
	}
	fclose(fd);

	return;
}

/*
 * Print the whole windows data into stdout.
 * The format is a data vector per line, each line containing
 * the number of metrics.
 */
void print_windows(double ***wins)
{
	unsigned int i, j, k;

	printf("Printing metric values:\n\n");
	for (i=0; i < num_windows; i++) {
		for (k=0; k < num_points; k++) {
			for (j=0; j < num_metrics; j++) {
				printf("%f",wins[i][j][k]);

				if (j != (num_metrics-1))
					printf(",");
				else
					printf("\n");
			}
		}
	}

	return;
}

/*
 * Calculate correlation coefficients vectors.
 * Input 'dim' is the dimensions of them (i.e., (n choose num_metrics)).
 */
void calc_ccvs(double **ccv, double ***win)
{
	unsigned int i,j,k,w;

#ifdef WITH_OPENMP
#pragma omp parallel for private(i,j,k,w) shared(num_windows,num_points,num_metrics,win,ccv)
#endif
	for (w=0; w < num_windows; w++) {
		vector<double> mu(num_metrics, 0);
		vector<double> sigma(num_metrics, 0);

		// Calculate mu and sigma for each series
		for (i=0; i < num_metrics; i++) {
			for (j=0; j < num_points; j++)
				mu[i] = mu[i] + win[w][i][j];
			mu[i] = mu[i] / (double)(num_points);

			for (j=0; j < num_points; j++)
				sigma[i] = sigma[i] + pow((double)(win[w][i][j] - mu[i]), 2);
			sigma[i] = sigma[i] / (double)(num_points - 1);
			sigma[i] = sqrt(sigma[i]);
		}

		// Calculate standard score for each value
		vector< vector<double> > std_score(num_metrics, vector<double>(num_points));
		for (i=0; i < num_metrics; i++) {
			for (j=0; j < num_points; j++) {
				if (sigma[i] > 0) {
					std_score[i][j] = (double)(win[w][i][j] - mu[i]) / sigma[i];
				} else {
					std_score[i][j] = 0;
				}
			}
		}

		// Calculate the correlation-coefficient vector
		unsigned int count=0;
		for (i=0; i < (num_metrics-1); i++) {
			for (j=i+1; j < num_metrics; j++) {
				double tmp = 0;
				for (k=0; k < std_score[i].size(); k++) {
					tmp = tmp + std_score[i][k] * std_score[j][k];
				}
				ccv[w][count] = tmp/(double)(std_score[i].size() - 1);
				count++;
			}
		}
	}

	return;
}

vector< vector<double> > calc_ccvs_from_data(double ***data)
{
	unsigned int i,j,k,w,c;

	vector< vector<double> > ccvs;

	unsigned int num_workloads = 3;
	//unsigned int num_values = 26640;
	unsigned int window_size = 100;
	unsigned int num_ccvs = 266; // ceiling(num_values / windows_size)

	for (w=0; w < num_workloads; w++) {
		vector< vector<double> > block(num_metrics, vector<double>(window_size)); // block[i][j], i: metric, j: value

		for (c=0; c < num_ccvs; c++) {

			/* Prepare block of data */
			for (i=0; i < num_metrics; i++)
				for (j=0; j < window_size; j++)
					block[i][j] = data[w][i][(j * num_ccvs) + j];

			vector<double> mu(num_metrics, 0);
			vector<double> sigma(num_metrics, 0);
			// Calculate mu and sigma for each series
			for (i=0; i < num_metrics; i++) {
				for (j=0; j < window_size; j++)
					mu[i] = mu[i] + block[i][j];
				mu[i] = mu[i] / (double)(num_points);

				for (j=0; j < window_size; j++)
					sigma[i] = sigma[i] + pow((double)(block[i][j] - mu[i]), 2);
				sigma[i] = sigma[i] / (double)(window_size - 1);
				sigma[i] = sqrt(sigma[i]);
			}

			// Calculate standard score for each value
			vector< vector<double> > std_score(num_metrics, vector<double>(window_size));
			for (i=0; i < num_metrics; i++) {
				for (j=0; j < window_size; j++) {
					if (sigma[i] > 0) {
						std_score[i][j] = (double)(block[i][j] - mu[i]) / sigma[i];
					} else {
						std_score[i][j] = 0;
					}
				}
			}

			// Calculate the correlation-coefficient vector
			vector<double> vect;
			for (i=0; i < (num_metrics-1); i++) {
				for (j=i+1; j < num_metrics; j++) {
					double tmp = 0;
					for (k=0; k < std_score[i].size(); k++) {
						tmp = tmp + std_score[i][k] * std_score[j][k];
					}
					//ccv[c][count] = tmp/(double)(std_score[i].size() - 1);
					vect.push_back(tmp/(double)(std_score[i].size() - 1));
				}
			}
			ccvs.push_back(vect);
		}
	}

	return ccvs;
}

/*
 * Determine what CCVs are constants.
 * Print results into a file.
 */
void check_const_ccvs(double **ccv, unsigned int *const_ccvs, double *last_values, unsigned int dim, const string &path)
{
	unsigned int i,j;

	for (i=0; i < dim; i++) {
		unsigned int is_const = 1;

		double value = ccv[0][i];
		for (j=1; j < num_windows; j++) {
			last_values[i] = ccv[j][i];
			if (value != ccv[j][i]) {
				is_const = 0;
				break;
			}
			value = ccv[j][i];
		}

		const_ccvs[i] = is_const;
	}

	/* Print constant CCVs into a file */
	FILE *fd;
	string filename = path + string("/constant_ccvs.dat");
	fd = fopen(filename.c_str(), "w");
	if (fd == NULL) {
		cerr << "Could not open file " + filename << endl;
	}

	fputs("#metric_pair,id,[0|1],value\n",fd);
	unsigned int count=0;
	char buffer[100];
	for (i=0; i < (num_metrics-1); i++) {
		for (j=i+1; j < num_metrics; j++) {
			buffer[0] = '\0';
			sprintf(buffer, "(%d|%d),%d,%d,%e\n",(i+1), (j+1), count, const_ccvs[count], last_values[count]);
			fputs(buffer,fd);
			count++;
		}
	}
	fclose(fd);

	/* Print CCVs into a file */
	filename = path + string("/corr_coeff_vectors.dat");
	fd = fopen(filename.c_str(), "w");
	if (fd == NULL) {
		cerr << "Could not open file " + filename << endl;
	}

	for (i=0; i < num_windows; i++) {
		for (j=0; j < dim; j++) {
			buffer[0] = '\0';
			sprintf(buffer, "%e",ccv[i][j]);
			fputs(buffer,fd);

			if (j != (dim-1)) fputs(",",fd);
			else fputs("\n",fd);
		}
	}
	fclose(fd);

	return;
}

void save_ccvs(const vector< vector<double> > ccvs, const string &path)
{

	FILE *fd;
	string filename = path + string("/ccvs_db.dat");
	fd = fopen(filename.c_str(), "w");
	if (fd == NULL) {
		cerr << "Could not open file " + filename << endl;
	}

	char buffer[100];
	unsigned int i,j;
	for (i=0; i < ccvs.size(); i++) {
		for (j=0; j < ccvs[i].size(); j++) {
			buffer[0] = '\0';
			sprintf(buffer, "%e",ccvs[i][j]);
			fputs(buffer,fd);

			if (j != (ccvs[i].size()-1)) fputs(",",fd);
			else fputs("\n",fd);
		}
	}
	fclose(fd);

	return;
}


int main(int argc, char *argv[])
{
	/* Program usage */
	if (argc < 2) {
		cerr << "Incorrect number of arguments: \n\t Usage: ./train_corr_coeff_vectors (win_directory)" << endl;
		exit(-1);
	}
	string path = string(argv[1]);

	/* Look for initial parameters (num_metrics, etc...) */
	set_params(path);

	/*
	 * Windows structure:
	 * windows[x][y][z]:
	 * 	x: number of windows
	 * 	y: number of metrics
	 * 	z: number of data points
	 */
	double ***wins = init_windows();

	/* Load data into windows structure */
	cout << "Loading windows...";
	load_windows(wins, path);
	cout << "done!" << endl;

	/* Load original data */
	double ***data = init_data();
	load_original_data(data);

	/* Check for constant metrics and build file */
	unsigned int *const_metrics = (unsigned int *) xmalloc(sizeof(unsigned int) * num_metrics);
	double *last_values = (double *) xmalloc(sizeof(double) * num_metrics);
	cout << "Checking for constant metrics...";
	check_const_metrics(const_metrics, last_values, wins, path);
	cout << "done!" << endl;

	/* Build correlation-coefficients vector (CCV) */
	unsigned int num_ccvs = n_choose_k(num_metrics, 2);
	double **ccv; // ccv data-structure

	/* Allocate space for CCVs */
	ccv = (double **) xmalloc(sizeof(double *) * num_windows);
	for (unsigned int i=0; i < num_windows; i++)
		ccv[i] = (double *) xmalloc(sizeof(double) * num_ccvs);

	/* Calculate CCVs */
	cout << "Calculating CCVs...";
	calc_ccvs(ccv, wins);
	cout << "done!" << endl;

	/* Check for constant CCVs */
	unsigned int *const_ccvs = (unsigned int *) xmalloc(sizeof(unsigned int) * num_ccvs);
	double *last_ccv_values = (double *) xmalloc(sizeof(double) * num_ccvs);
	check_const_ccvs(ccv, const_ccvs, last_ccv_values, num_ccvs, path);

	/* Calculate sliding-window CCVs */
	vector< vector<double> > real_ccvs = calc_ccvs_from_data(data);
	save_ccvs(real_ccvs, path);

	//cout << "Path: " << path << " Params: " << num_metrics << " " << num_windows <<  " " << num_points << endl;

	//print_windows(wins);

	return 0;
}
