#!/usr/bin/perl
use warnings;
use strict;

###### Variables ########
my $numRuns = 17;
my $abnoral_db_path="../database/abnormal/fault_10";
my $experiment_output_path="./data";
my %injection_sample; # key is the run-id
my %failure_sample;
#########################

# Load Failure point table
my $filename = $abnoral_db_path."/failure_points_table.dat";
open FILE, $filename or die $!;
while (my $line = <FILE>) {
	chomp($line);
	
	# Omit lines strating with '#'
	if ($line =~ m/^\#/g) {
		next;
	}
	
	my @tmp = split(/\,/, $line);
	
	# Save injection and failure samples in hash table
	$injection_sample{$tmp[0]} = $tmp[1];
	$failure_sample{$tmp[0]} = $tmp[2];
}
close(FILE);

### No Forecasting Metrics ###
my $nf_false_positive = 0;
my $nf_true_positive = 0;
my $nf_false_negative = 0;
my @nf_look_ahead;

### Forecasting Metrics ###
my $f_false_positive = 0;
my $f_true_positive = 0;
my $f_false_negative = 0;
my @f_look_ahead;

my @f_look_ahead_preceded;
my @f_look_ahead_preceded_percent_of_diff; # percent of the difference with respect to no-forcast time 

for (my $i=1; $i <= $numRuns; $i++) {
	if ($failure_sample{$i} == 0) {
		next;
	}
	
	my $fn_bound = $failure_sample{$i} + 50; # 50 5-sec samples
	
	############### Run No-Forecasting ###############
	system "./run_producer.sh $i $abnoral_db_path &";
	system "./run_consumer.sh $i 0";
	
	# Find error sample
	my $error_sample = `./find_error_sample.pl $experiment_output_path/out_$i.dat`;
	chomp($error_sample);
	
	my $nf_had_look_ahead = 0;
	my $nf_last_la = 0;
	
	if ($error_sample == 0) { # false negative
		$nf_false_negative = $nf_false_negative + 1;
	} elsif ($error_sample <= $injection_sample{$i}) { # flase positive
		$nf_false_positive = $nf_false_positive + 1;
	} elsif ($error_sample > $injection_sample{$i} && $error_sample <= $failure_sample{$i}) { # true positive
		$nf_true_positive = $nf_true_positive + 1;
		
		# Look-ahead
		my $diff_samples = $failure_sample{$i} - $error_sample;
		push(@nf_look_ahead, $diff_samples);
		$nf_last_la = $diff_samples;
		$nf_had_look_ahead = 1;
	} elsif ($error_sample > $failure_sample{$i} && $error_sample <= $fn_bound) { # true positive
		$nf_true_positive = $nf_true_positive + 1;
	} elsif ($error_sample > $fn_bound) { # false negative
		$nf_false_negative = $nf_false_negative + 1;
	}
		
	sleep(1);
	
	############### Run Forecasting ###############
	system "./run_producer.sh $i $abnoral_db_path &";
	system "./run_consumer.sh $i 1";
	
	# Find error sample
	$error_sample = `./find_error_sample.pl $experiment_output_path/out_$i.dat`;
	chomp($error_sample);
	
	my $f_had_look_ahead = 0;
	my $f_last_la = 0;
	
	if ($error_sample == 0) { # false negative
		$f_false_negative = $nf_false_negative + 1;
	} elsif ($error_sample <= $injection_sample{$i}) { # flase positive
		$f_false_positive = $f_false_positive + 1;
	} elsif ($error_sample > $injection_sample{$i} && $error_sample <= $failure_sample{$i}) { # true positive
		$f_true_positive = $f_true_positive + 1;
		
		# Look-ahead
		my $diff_samples = $failure_sample{$i} - $error_sample;
		push(@f_look_ahead, $diff_samples);
		$f_last_la = $diff_samples;
		$f_had_look_ahead = 1;
	} elsif ($error_sample > $failure_sample{$i} && $error_sample <= $fn_bound) { # true positive
		$f_true_positive = $f_true_positive + 1;
	} elsif ($error_sample > $fn_bound) { # false negative
		$f_false_negative = $nf_false_negative + 1;
	}
	
	##################################################
	# Check if look-ahead when forecasting precedes 
	# the one with no-forecasting
	##################################################
	if ($nf_had_look_ahead == 1 && $f_had_look_ahead == 1) { # Only if in both cases there was look-ahead time
		if ($f_last_la > $nf_last_la) {
			my $diff = $f_last_la - $nf_last_la;
			push(@f_look_ahead_preceded, $diff);
			
			my $per = $diff / $nf_last_la;
			push(@f_look_ahead_preceded_percent_of_diff, $per);
		}
	}
}

############################
##### Stats generation #####
############################

### No forecasting ###
my $nf_precision = $nf_true_positive / ($nf_true_positive + $nf_false_positive);
my $nf_recall = $nf_true_positive / ($nf_true_positive + $nf_false_negative);

# Get average look-ahead
my $mu_nf_look_ahead = 0;
if (@nf_look_ahead) {
	for (my $i=0; $i < @nf_look_ahead; $i++) {
		$mu_nf_look_ahead = $mu_nf_look_ahead + $nf_look_ahead[$i];
	}
	$mu_nf_look_ahead = $mu_nf_look_ahead / @nf_look_ahead;
}


### Forecasting ###
my $f_precision = $f_true_positive / ($f_true_positive + $f_false_positive);
my $f_recall = $f_true_positive / ($f_true_positive + $f_false_negative);

# Get average look-ahead
my $mu_f_look_ahead = 0;
if (@f_look_ahead) {
	for (my $i=0; $i < @f_look_ahead; $i++) {
		$mu_f_look_ahead = $mu_f_look_ahead + $f_look_ahead[$i];
	}
	$mu_f_look_ahead = $mu_f_look_ahead / @f_look_ahead;
}


### Forecast Look-ahead time precedes no-forecast time
my $f_la_preceded_percentage = 0;
if (@f_look_ahead) {
	$f_la_preceded_percentage = @f_look_ahead_preceded / @f_look_ahead;
}

my $f_la_preceded_mu = 0;
if (@f_look_ahead_preceded_percent_of_diff) {
	for (my $i=0; $i < @f_look_ahead_preceded_percent_of_diff; $i++) {
		$f_la_preceded_mu = $f_la_preceded_mu + $f_look_ahead_preceded_percent_of_diff[$i];
	}
	$f_la_preceded_mu = $f_la_preceded_mu / @f_look_ahead_preceded_percent_of_diff;
}

print "# Format: (NF=no forecast, F=forecast, la=look-ahead-avg\n";
print "NF_tp,NF_fp,NF_fn,NF_recall,NF_precision,NF_la,F_tp,F_fp,F_fn,F_recall,F_precision,F_la,F_preceded_per,F_preceded_diff_per\n";

print "$nf_true_positive,$nf_false_positive,$nf_false_negative,$nf_recall,$nf_precision,$mu_nf_look_ahead,";
print "$f_true_positive,$f_false_positive,$f_false_negative,$f_recall,$f_precision,$mu_f_look_ahead,";
print "$f_la_preceded_percentage,$f_la_preceded_mu\n";


