/*
 * arima_tools.cpp
 *
 *  Created on: Jun 9, 2010
 *      Author: Ignacio Laguna
 *     Contact: ilaguna@purdue.edu
 */
#include "arima_tools.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <vector>
#include <iostream>

using namespace std;

/*
 * -------------------
 * Private prototypes
 * -------------------
 */

/*
 * ------------------
 * Public functions
 * ------------------
 */
void ARIMA_predict(ARIMAModel *model, unsigned int num, const vector<double> & values, Predictions *pred)
{
	//Predictions *pred = NULL;
	vector<double> errors;
	unsigned int ma_params = model->ma_params.size();
	unsigned int values_size;
	vector<double> values_trans;
	Series dseries;

	/*
	 * Differentiate
	 */
	if (model->d > 0) {
		//cout << model->d << " ...differencing \n";
		dseries.diff(values, values_trans, model->d);
	} else {
		values_trans = values;
	}

	values_size = values_trans.size();

	/*
	 * Initialize error values;
	 */
	for (unsigned int i=0; i < ma_params; i++) {
		errors.push_back(0.0);
	}

	/*
	 * Create whole window of values: observed + predicted
	 */
	vector<double> win;
	for (unsigned int i=0; i < values_size; i++) {
		win.push_back(values_trans[i]);
	}
	for (unsigned int i=0; i < num; i++) {
		win.push_back(0.0);
	}

	if (ma_params > 0) {
	//	cout << "Has MA" << endl;
		/*
		 * Make predictions when q > 0,
		 * when we need the previous error values.
		 */
		unsigned int start = model->ar_params.size();
		for (unsigned int i=start; i < win.size(); i++) {

			double new_val = 0;
			double error_now = 0;
			double phi_factor = 1.0;

			// Do the AR part
			if (model->ar_params.size() > 0) {
				//cout << "Do AR part" << endl;
				for (unsigned int j=0; j < model->ar_params.size(); j++) {
					new_val = new_val + model->ar_params[j] * win[i-1-j];
					phi_factor = phi_factor - model->ar_params[j];
				}
			}

			// Do the MA part
			for (unsigned int j=0; j < ma_params; j++) {
				new_val = new_val + model->ma_params[j] * errors[ma_params-1-j];
			}

			// Add intersect
			new_val = new_val + model->intercept * phi_factor;

			// Update predictions
			if (i >= values_size) {
				win[i] = new_val;
				error_now = 0;
			} else {
				error_now = win[i] - new_val;
			}

			// Update errors
			for (unsigned int i=0; i < ma_params-1; i++) {
				errors[i] = errors[i+1];
			}
			errors[ma_params-1] = error_now;
			//errors[ma_params-1] = 0;
		}
	} else if (model->ma_params.size() == 0) {
		//cout << "NO MA" << endl;
		/*
		 * Here q =0, so we don't need the previous error values
		 */
		//unsigned int start = values.size();
		for (unsigned int i=values_size; i < win.size(); i++) {
			double new_val = 0;
			double phi_factor = 1.0;

			// Do the AR part
			if (model->ar_params.size() > 0) {
				for (unsigned int j=0; j < model->ar_params.size(); j++) {
					new_val = new_val + model->ar_params[j] * win[i-1-j];
					phi_factor = phi_factor - model->ar_params[j];
				}
			}

			// Add intersect
			new_val = new_val + model->intercept * phi_factor;

			// Update predictions
			win[i] = new_val;
		}
	}

	/*for (unsigned int i=0; i < win.size(); i++) {
		cout << "__________Win " << i << " " << win[i] << endl;
	}*/

	/*
	 * Update prediction structure
	 */
	pred->values.clear();
	pred->error.clear();
	for (unsigned int i=0; i < num; i++) {
		pred->values.push_back(win[values_size+i]);
		pred->error.push_back(1.96 * sqrt(model->sigma_2));
	}

	/*
	 * Undo difference only if d > 0
	 */
	if (model->d > 0) {
		//cout << "...undoing difference\n";
		vector<double> tmp = pred->values;
		dseries.undo_diff(tmp, pred->values);
	}

	return;
}


/*void ARIMA_best_model(vector<double> values, ARIMAModel *model)
{

	return;
}*/

void Series::diff(const vector<double> &old_series, vector<double> &new_series, unsigned int d)
{

	unsigned int s = old_series.size();
	diff_d = d;

	if (d >= s) {
		printf("[ARIMA_TOOLS] Cannot differentiate series\n");
		fflush(stdout);
		exit(-1);
	}

	/*
	 * Save original window
	 */
	vector<double> tmp;
	for (unsigned int i=0; i < s; i++)
		tmp.push_back(old_series[i]);
	dseries.win.push_back(tmp);

	/*
	 * Differentiate d times
	 */
	for (unsigned int i=0; i < d; i++) {
		tmp.clear();
		for (unsigned int j=0; j < (dseries.win[i].size() - 1); j++) {
			tmp.push_back(dseries.win[i][j+1] - dseries.win[i][j]);
		}
		dseries.win.push_back(tmp);
	}

	new_series = dseries.win[d];

	return;
}

void Series::print(void)
{
	cout << "Printing series ..." << endl;
	for (unsigned int i=0; i < dseries.win.size(); i++) {
		for (unsigned int j=0; j < dseries.win[i].size(); j++) {
			cout << dseries.win[i][j] << " ";
		}
		cout << endl;
	}
	return;
}

void Series::undo_diff(vector<double> trans_predict, vector<double> &real_predict)
{
	if (diff_d <= 0) {
		printf("[ARIMA_TOOLS] Cannot undo differentiated series\n");
		fflush(stdout);
		exit(-1);
	}

	//cout << "Diff value: " << diff_d << endl;

	unsigned int size = trans_predict.size();

	real_predict.clear();
	for (unsigned int i=0; i < size; i++)
		real_predict.push_back(0);

	/*
	 * Undo d-differentiated series
	 */
	for (unsigned int i=0; i < diff_d; i++) {
		unsigned int s = dseries.win.size();

		for (unsigned int k=0; k < size; k++) {
			double prev = 0;
			if (k == 0) prev = dseries.win[s-i-2].back();
			else prev = real_predict[k-1];

			real_predict[k] = prev + trans_predict[k];
		}

		trans_predict = real_predict;
	}

	return;
}

/*
 * -------------------
 * Private functions
 * -------------------
 */
