/*
 * data_producer.C
 *
 *  Created on: Jan 3, 2011
 *      Author: Ignacio Laguna
 *     Contact: ilaguna@purdue.edu
 */

#include "detection_engine.h"

#include <assert.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/msg.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace std;

/* Global variables */
unsigned int num_metrics = 0;
unsigned int num_points = 0;

void set_global_variables(void)
{
	char *var;

	/* Get number of metrics */
	if ((var = getenv("ERROR_DETECT_NUM_METRICS")) != NULL) {
		num_metrics = static_cast<unsigned int>(atoi(var));
	} else {
		cerr << "[data_producer] Error in set_global_variables(): could not find ERROR_DETECT_NUM_METRICS" << endl;
		cerr.flush();
		exit(-1);
	}

	/* Get number of data points */
	if ((var = getenv("ERROR_DETECT_NUM_POINTS")) != NULL) {
		num_points = static_cast<unsigned int>(atoi(var));
	} else {
		cerr << "[data_producer] Error in set_global_variables(): could not find ERROR_DETECT_NUM_POINTS" << endl;
		cerr.flush();
		exit(-1);
	}

	return;
}

void print_usage(void)
{
	cout << "Usage:\n./data_producer (data_file)" << endl;
	cout.flush();
	exit(-1);
}

/*
 * Error-checking malloc
 */
void * xmalloc (size_t size)
{
  register void *value = malloc(size);
  if (value == 0) {
    cerr << "virtual memory exhausted" << endl;
    exit(-1);
  }
  return value;
}

void Tokenize(const string& str, vector<string>& tokens, const string& delimiters = " ")
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

/*
 * Delete the last character of a string.
 */
void del_last_char(char *line)
{
	size_t n = strlen(line);
	line[n-1] = '\0';
	return;
}

/*
 * Clean message structure
 */
inline
void clean_msg_structure(MsgType *msg)
{
	for (unsigned int i=0; i < MSG_SIZE; i++)
		msg->metrics[i] = 0;

	return;
}

int
main (int argc, char *argv[])
{
	if (argc != 2) print_usage();

	/* Set global variables */
	set_global_variables();

	char *filename;
	filename = argv[1];
	FILE *file;

	/*
	 * ---------
	 * Open file
	 * ---------
	 */
	file = fopen(filename, "r");
	if (file == (FILE *)0) {
		cerr << "[data_producer]: Could not open data file." << endl;
		cerr.flush();
		exit(-1);
	}

	/*
	 * Allocate msg buffer
	 */
	MsgType msg; // message structure
	clean_msg_structure(&msg);


	/*
	 * --------------------------------------
	 * Set message queue for detection engine
	 * --------------------------------------
	 */
	int qid, ret;
	struct msqid_ds buf;
	qid = msgget(QUEUE_ID, 0666 | IPC_CREAT); // get queue id
	if (qid >=0) {
		cout << "[data_producer]: Msg queue created, id=" << qid << endl;
		cout.flush();

		// Configure queue size
		ret = msgctl(qid, IPC_STAT, &buf);
		buf.msg_qbytes = 131072;
		if (msgctl(qid, IPC_SET, &buf) == 0) {
			cout << "[data_producer]: Queue configured..." << endl;
			cout.flush();
		}
	}

	/*
	 * -------------------------------
	 * Loop to send messages from file
	 * -------------------------------
	 */
	ssize_t r;
	size_t n = 5000;
	char *lineptr = (char *)malloc(sizeof(char) * n);
	r = getline(&lineptr, &n, file);
	if (r == -1) {
		cerr << "[data_producer]: Error reading line from file." << endl;
		cerr.flush();
		exit(-1);
	}

	while (r > 0) {
		del_last_char(lineptr);
		string line_str(lineptr);
		vector<string> tokens;
		Tokenize(line_str,tokens,",");

		/* Fill message with values from a line of the file */
		for (size_t i=0; i < tokens.size(); i++)
			msg.metrics[i] = atof(tokens[i].c_str());

		/* Send message */
		msg.type = 1L;
		if (msgsnd(qid, (struct msgbuf *)&msg, sizeof(msg), 0) == -1) {
			cerr << "[data_producer]: failed to send message" << endl;
			cerr.flush();
			exit(-1);
		}

		/* Get line */
		r = getline(&lineptr, &n, file);
	}

	/*
	 * ----------
	 * Close file
	 * ----------
	 */
	cout << "[data_producer]: Done with file" << endl;
	fclose(file);

	/*
	 * --------------------------
	 * Remove queue if it's empty
	 * --------------------------
	 */
	cout << "[data_producer]: Removing queue...";
	cout.flush();
	msgctl(qid, IPC_STAT, &buf);
	while ((int)(buf.msg_qnum) > 0) {
		msgctl(qid, IPC_STAT, &buf);
	}
	if (msgctl(qid, IPC_RMID, NULL) == -1) {
		cerr << "[data_producer]: failed removing queue" << endl;
		cerr.flush();
		exit(-1);
	}
	cout << "done!" << endl;

	return 0;
}
