/*
 * detection_engine.C
 *
 *  Created on: Jan 3, 2011
 *      Author: Ignacio Laguna
 *     Contact: ilaguna@purdue.edu
 */

#include "utilities.h"
#include "detection_engine.h"
#include "arima_tools.h"
#include "nearest_neighbor.h"
#include "timing.h"
#include "set_config.h"
#include "data_loading.h"
#include "correlation_coeff.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/msg.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include <math.h>

#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <map>

using namespace std;

/*------------------------------------------------------------*/
/* Global variables                                           */
/*------------------------------------------------------------*/

/* Configuration parameters */
Config config;

/* ARIMA models database */
vector<Models> modelsBase;

/* Database root directory */
string database_path;

/* Window of observations */
vector< vector<double> > window;

struct timeval curr_time;

/* Filtered constant dimensions (or metrics) */
vector<unsigned int> const_metrics;
vector<double> const_values;

/*
 * Map of CCVs valid dimensions and their corresponding pair of metrics.
 * A CCV dimension is valid only is both of the two metrics are non-constant.
 */
map<unsigned int, MetricPair> validCCVDimensions;

/*
 * Nearest-Neighbor hyper-sphere.
 * Dimensions of h_sphere[x][y]:
 * x: window number (or sample id)
 * y: correlation-coefficient vector
 */
vector< vector<double> > h_sphere;

/* Time measurements */
map<string, vector<double> > time_traces;

/* Metric names */
map<unsigned int, string> metricNames;

/* Standard deviations (from training set) */
vector<double> standardDeviation;

/* KNN normal distance threshold */
double knn_distance_mean = 2.4548;
double knn_distance_std = 1.6025;

/* True when an error has been flagged */
bool error_flagged = false;

/* Burn-in maximum sample. We do not check for errors before this sample */
int burnin_threshold = 160;

/* Current sample */
int sample;
int beakingPointSample = 807;

/*-----------------------------------------------------------------*/
/* Private Functions                                               */
/*-----------------------------------------------------------------*/

inline
void print_usage(void)
{
	cout << "Usage:\n./detection_engine (database dir)" << endl;
	cout.flush();
	exit(EXIT_FAILURE);
}

void printDesiredCorrelations(const vector <double> &ccv) {

	vector<unsigned int> desiredMetrics;
	desiredMetrics.push_back(22); // Metric 23 is tomcat's memory

	printf("DESIRED_METRIC_NAMES,");
	for (size_t i=0; i < ccv.size(); i++) {
		if (isCCVDimensionWanted(i, desiredMetrics)) {
			MetricPair mpair = validCCVDimensions[static_cast<unsigned int>(i)];

			string name_x = metricNames[mpair.x];
			string name_y = metricNames[mpair.y];
			printf("(%s|%s),",name_x.c_str(), name_y.c_str());
		}
	}
	printf("\n");

	printf("DESIRED_CCS,");
	for (size_t i=0; i < ccv.size(); i++) {
		if (isCCVDimensionWanted(i, desiredMetrics)) {
			printf("%f,",ccv[i]);
		}
	}
	printf("\n");
}

/*
 * Filter constant metrics.
 * This function should be used right after getting an
 * observation vector from the producer process.

 * If a constant metric is different from its
 * regular 'constant value', an error is flagged.
 */
vector<double> filter_const_metrics(const vector<double> &obs_vector,
		bool &error)
{
	vector<double> ret;

	for (unsigned int i=0; i < obs_vector.size(); i++) {
		if (const_metrics[i] == 0) { // check if it's non-constant
			ret.push_back(obs_vector[i]);
		} else { // if it's constant, check its value
			if (const_values[i] == obs_vector[i])
				error = false;
			else
				error = true;
		}
	}

	return ret;
}

/*
 * Get observations vector from a queue-message
 */
vector<double> get_obs_vector(const MsgType &msg)
{
	vector<double> ret;

	for (unsigned int i=0; i < config.num_metrics; i++)
		ret.push_back(msg.metrics[i]);

	return ret;
}

/*
 * Updates the (global) window of observations.
 * Returns:
 * TRUE: if we have enough observations to do predictions
 * FALSE: otherwise
 *
 * The window is managed as a 'sliding window'. Once the
 * windows is equal or greater than a particular size,
 * we add points to the end and delete points at the beginning.
 */
bool update_window(const vector<double> &m)
{
	bool ret = false;

	if (window.size() >= config.win_size) {
		window.push_back(m);
		window.erase(window.begin());
		ret = true;
	} else {
		window.push_back(m);
	}

	return ret;
}

/*
 * Calculate Root-Mean-Square (RMS) of two vectors.
 */
double get_RMS(const vector<double> &x, const vector<double> &y)
{
	double ret = 0;

	if (x.size() == y.size()) {
		for (unsigned int i=0; i < x.size(); i++) {
			double tmp = (x[i] - y[i]);
			ret = ret + pow(tmp,2.0);
		}
		ret = ret/(x.size());
		ret = sqrt(ret);
	} else {
		handle_error("in get_RMS(): vectors are not of the same size");
	}

	return ret;
}

double get_error_based_on_mean(const vector<double> &x, const vector<double> &y)
{
	double ret = 0;

	if (x.size() == y.size()) {
		double mean_x = 0;
		for (size_t i=0; i < x.size(); i++)
			mean_x += x[i];
		mean_x = mean_x / x.size();

		double mean_y = 0;
		for (size_t i=0; i < y.size(); i++)
			mean_y += y[i];
		mean_y = mean_y / y.size();

		ret = fabs(mean_x - mean_y);
	} else {
		handle_error("in get_error_based_on_mean(): "
				"vectors are not of the same size");
	}

	return ret;
}

/*
 * Check that the ARIMA models is good to be used in predictions.
 * For example, an ARIMA(0,0,0) can not be used.
 */
bool good_model(const ARIMAModel &model)
{
	bool ret = true;

	if (model.ar_params.size() > 0 && model.ma_params.size() > 0) {
		if (model.d == 0 &&
				model.ar_params[0] == 0 &&
				model.ma_params[0] == 0) {
			ret = false;
		}
	}

	return ret;
}

void print_ARIMA_prediction(const vector<double> &obs, unsigned int metric)
{

	Models models = modelsBase[metric];
	unsigned int numPredictions = 10;

	// Iterate over all models for this metric
	for (size_t i = 0; i < models.mod.size(); ++i) {
		Predictions pred;
		ARIMA_predict(&(models.mod[i]), numPredictions, obs, &pred);

		// Save predictions in files
		char metric_id[10];
		metric_id[0] = '\0';
		sprintf(metric_id, "%d", static_cast<int>(i));

		FILE *file_ptr;
		string file_name= string("./data/model_") +
				string(metric_id) + string(".dat");
		if (!(file_ptr = fopen(file_name.c_str(), "w"))) {
			string msg("Error opening file: " + file_name);
			handle_error(msg.c_str());
		}

		// Save observations
		for (size_t j=0; j < obs.size(); ++j) {
			char buffer[50];
			buffer[0] = '\0';
			sprintf(buffer, "%e,", obs[j]);
			fputs(buffer,file_ptr);
		}

		// Save predictions
		for (size_t j=0; j < pred.values.size(); ++j) {
			char buffer[50];
			buffer[0] = '\0';
			sprintf(buffer, "%e,", pred.values[j]);
			fputs(buffer,file_ptr);
		}

		fputs("\n",file_ptr);

		fclose(file_ptr);
	}
}

bool window_is_constant(unsigned int metric) {
	bool ret = true;

	double lastValue = window[0][metric];
	for (unsigned int k=1; k < window.size(); k++) {
		if (window[k][metric] != lastValue) {
			ret = false;
			break;
		}
		lastValue = window[k][metric];
	}

	return ret;
}

/*
 * This function returns a vector of predictions.
 * It selects the best model to predict each metric.

 * IMPORTANT: the vector ret[i][j] that is returned is as follows:
 * i: metric
 * j: predictions of metric i
 */
vector< vector<double> > get_predictions(void)
{
	vector< vector <double> > ret;
	Predictions predict;

	/*
	 * Loop to select best model
	 */
	unsigned int metrics = modelsBase.size();
	// iterate over each metric
	for (unsigned int i = 0; i < metrics; i++) {

		//get_time_start();
		// Do not find model if window is filled with constant values
		if (window_is_constant(i)) {
			vector<double> pred(config.predict_size);
			// Fill predictions with constant value
			for (size_t j = 0; j < pred.size(); j++)
				pred[j] = window[0][i];

			ret.push_back(pred);
			continue;
		}

		Models m = modelsBase[i];
		unsigned int num_models = m.mod.size();

		// get observations for a column (i.e., for a metric)
		vector<double> obs;
		vector<double> ground_th;
		/*unsigned int lim = (unsigned int)(config.win_size/2);
		for (unsigned int k=0; k < lim; k++) {
			obs.push_back(window[k][i]);
			ground_th.push_back(window[k+lim][i]);
		}*/

		// percentage of data to take as observations
		double per_observations = 0.9;
		unsigned int num_obs = static_cast<unsigned int>(
				static_cast<double>(config.win_size) * per_observations);
		unsigned int num_test_points = config.win_size - num_obs;
		unsigned int lim = num_test_points;

		// Fill observation and testing sets
		for (unsigned int k=0; k < num_obs; k++)
			obs.push_back(window[k][i]);
		for (unsigned int k=0; k < num_test_points; k++)
			ground_th.push_back(window[num_obs + k][i]);

		// get the complete window
		vector<double> win;
		for (unsigned int k=0; k < config.win_size; k++)
			win.push_back(window[k][i]);

		double error = -1;
		Predictions pred;
		unsigned int best_model_pointer = 0;
		// iterate over each model
		for (unsigned int j = 0; j < num_models; j++) {
			ARIMA_predict(&(modelsBase[i].mod[j]), lim, obs, &pred);
			double tmp = get_RMS(pred.values, ground_th);
			//double tmp = get_error_based_on_mean(pred.values, ground_th);

			/*
			 * Selects model with smallest RMS
			 */
			if (error == -1) {
				best_model_pointer = j;
				error = tmp;
			} else if (tmp < error) {
				best_model_pointer = j;
				error = tmp;
			}
		}
		//get_time_stop("FIND_BEST_MODELS_PER_METRIC");

		// Print predictions for each model
		/*if ((sample == beakingPointSample) && (i == 0)) {
			print_ARIMA_prediction(obs,i);
		}*/

		/*
		 * Now, make predictions using the best model,
		 * and using the complete window of observations
		 */
		//get_time_start();
		vector<double> tmp;
		if (!good_model(modelsBase[i].mod[best_model_pointer])) {

			//cout << "NOT GOOD MODEL! Metric: " << i << endl;

			// if this the case, then use the last n observations
			// as the predictions
			for (unsigned int k=0; k < config.predict_size; k++)
				tmp.push_back(win[win.size() - k - 1]);

		} else {
			ARIMA_predict(&(modelsBase[i].mod[best_model_pointer]),
					config.predict_size, win, &pred);
			for (unsigned int k=0; k < config.predict_size; k++)
				tmp.push_back(pred.values[k]);
		}

		ret.push_back(tmp);
		//get_time_stop("PERFORM_FORECASTS_PER_METRIC");
	}

	return ret;
}

/*
 * IMPORTANT: the vector predict[i][j] that is given as input is as follows:
 * i: metric
 * j: predictions of metric i
 */
vector< vector<double> > unify_observations_and_predictions(
		const vector< vector<double> > &predict) {

	//unsigned int metrics = modelsBase.size();
	unsigned int metrics = window[0].size();
	vector< vector<double> > data;

	/*
	 * Unify observations and predictions.
	 * NOTICE: predict and window vectors have different format.
	 */
	for (unsigned int m=0; m < metrics; m++) {
		vector<double> tmp;

		for (unsigned int i=0; i < config.win_size; i++)
			tmp.push_back(window[i][m]);

		if (config.use_predictions) {
			// Add predictions to temporal vector
			for (unsigned int i=0; i < predict[m].size(); i++) {
				tmp.push_back(predict[m][i]);
			}

			// Eliminate the beginning of the vector so that
			// the evaluated window is of the original size
			//tmp.erase(tmp.begin(), tmp.begin() + config.predict_size - 1);
		}

		data.push_back(tmp);
	}

	return data;
}

void print_observations_and_predictions(const vector< vector<double> > &data,
		int sample) {
	if (sample == beakingPointSample) {

		FILE *file_ptr;
		string file_name("./pred_metrics.dat");
		if (!(file_ptr = fopen(file_name.c_str(), "w"))) {
			string msg("Error opening file: " + file_name);
			handle_error(msg.c_str());
		}

		for (size_t i = 0; i < data[0].size(); ++i) {
			for (size_t j = 0; j < data.size(); ++j) {
				char buffer[50];
				buffer[0] = '\0';
				sprintf(buffer, "%e,", data[j][i]);
				fputs(buffer,file_ptr);
				//printf("%e,",data[j][i]);
			}
			//printf("\n");
			fputs("\n",file_ptr);
		}
		fclose(file_ptr);
	}
}

void print_ccv(const vector<double> &ccv) {

	FILE *file_ptr;
	string file_name("./sampled_ccvs.dat");
	if (!(file_ptr = fopen(file_name.c_str(), "a"))) {
		string msg("Error opening file: " + file_name);
		handle_error(msg.c_str());
	}

	for (size_t i = 0; i < ccv.size(); ++i) {
		char buffer[50];
		buffer[0] = '\0';
		sprintf(buffer, "%e,", ccv[i]);
		fputs(buffer,file_ptr);
	}
	fputs("\n",file_ptr);

	fclose(file_ptr);
}

/*
 * Clean message structure
 */
inline
void clean_msg_structure(MsgType *msg)
{
	for (unsigned int i=0; i < MSG_SIZE; i++)
		msg->metrics[i] = 0;

	return;
}

/* Did we pass the burn-in phase? */
bool afterBurninPhase(int sample) {
	bool ret = false;
	if (sample > burnin_threshold) {
		ret = true;
	}
	return ret;
}

/*
 * Flag error if it has not been flagged before.
 */
void flag_error_if_not_done_before(double distance, int sample) {
	if (afterBurninPhase(sample)) {
		if (!error_flagged) {
			double threshold = knn_distance_mean + 3.0*knn_distance_std;
			if (distance > threshold) {
				cout << "ERROR_DETECTION SAMPLE " << sample <<
						" DISTANCE " << distance << endl;
				error_flagged = true;
			}
		}
	}
}

/*-----------------------------------------------------------------*/
/* Main function                                                   */
/*-----------------------------------------------------------------*/
/*
 * It gets messages from a kernel message queue. A message is a vector of
 * observations. It prints out 'alarms' when an error is detected.
 */
int
main(int argc, char *argv[])
{
	if (argc != 2)
		print_usage();

	// Get input parameter
	database_path = string(argv[1]);

	/* Set configuration parameters */
	set_config_parameters();

	/* Set Queue variables */
	MsgType msg;
	clean_msg_structure(&msg);
	int qid, ret;

	/* Load flags to filter constant dimensions. */
	//cout << "[detection_engine]: Loading constant-dimension vector...";
	load_const_dims_vector();
	//cout << "done!" << endl;

	/* Load ARIMA models */
	//cout << "[detection_engine]: Loading ARIMA models...";
	//get_time_start();
	load_models();
	//get_time_stop("LOAD_ARIMA_MODELS");
	//cout << "done!" << endl;

	/* Load Hyper-sphere (for KNN) */
	//get_time_start();
	load_ccvs();
	//get_time_stop("LOAD_HYPER_SPHERE");

	/* Calculate KNN detection threshold */
	//calculate_knn_thresh();

	/* Load valid CCV-dimensions */
	loadValidCCVDimensions();

	/* Load metric names */
	load_metric_names();

	/* Load STDs */
	load_standard_deviations();

    	/* Get the message queue id */
    	qid = msgget(QUEUE_ID, 0);

    	if (qid >= 0) {
    		sample = 0;
    		while (true) {
    			/*
    			 * Receive message from the queue. If there is no message,
    			 * the process blocks until a message is available.
    			 * We receive two message types: 'tomcat' and 'mysql' vectors.
    			 */
    			ret = msgrcv(qid, (struct msgbuf *)&msg, sizeof(MsgType), 1, 0);

    			if (ret != -1) {
    				/*
    				 * The messages were received correctly.
    				 * Now we do the following steps here:
    				 * (1) Predict sequences,
    				 * (2) Get correlation coefficient vector
    				 * (3) Flag error
    				 */

        			/* Update sample count */
        			sample++;

    				/* Get vector of measurements */
    				//get_time_start();
    				vector<double> m =  get_obs_vector(msg);

    				/* Filter constant metrics */
    				bool error = false;
    				m = filter_const_metrics(m, error);
    				//get_time_stop("GET_OBS_AND_FILTER_CONSTS");

    				/* Check if we have enough measurements to do predictions */
    				if (update_window(m)) {
    					/* Make predictions */
    					//get_time_start();
    					vector< vector <double> > predict = get_predictions();
    					//get_time_stop("MAKE_PREDICTIONS");

    					/* Calculate correlation-coefficient vector (CCV) */
    					//get_time_start();
    					vector< vector<double> > data =
    							unify_observations_and_predictions(predict);
    					//get_time_stop("JOIN_OBS_AND_FORECASTS");
    					//print_observations_and_predictions(data,sample);
    					//get_time_start();
    					vector<double> ccv = pearson_coefficient_vector(data);
    					//vector<double> ccv = kl_coefficient_vector(data);
    					//vector<double> ccv = pearson_coefficient_vector_scaled(data);
    					print_ccv(ccv);
    					//get_time_stop("CALCULATE_CCV");

    					//printDesiredCorrelations(ccv);

    					/* Calculate CCV distance */
    					//get_time_start();
    					double dist = knn_ccv_distance(ccv, h_sphere);
    					//get_time_stop("KNN_DISTANCE");
    					printf ("%e\n", dist);
    					//cout << "KNN_DIST: " << dist << endl;

    					/* Detect error */
    					flag_error_if_not_done_before(dist, sample);
    				}

    			} else {
    	    		// Print statistics
    	    		print_times();

    	    		// Handle error
    				handle_error("[detection_engine]: failed to receive message"
    						"---If the other process terminated, "
    						"ignore this message");
    			}
    		}
    	} else {
    		// Handle error
    		char buffer[50];
    		buffer[0] = '\0';
    		sprintf(buffer, "%d", qid);
    		string msg = "[detection_engine]: failed to get message queue id. "
    				"Error_no=" + string(buffer);
    		handle_error(msg.c_str());
    	}

	return 0;
}
