/*
 * nearest_neighbor.C
 *
 *  Created on: Feb 9, 2011
 *      Author: Ignacio Laguna
 *     Contact: ilaguna@purdue.edu
 */

#include "nearest_neighbor.h"
#include "global.h"
#include "detection_engine.h"
#include "utilities.h"
#include <math.h>
#include <stdlib.h>
#include <time.h>

#include <omp.h>

#include <vector>
#include <map>

using namespace std;

bool isCCVDimensionWanted(unsigned int dim,
		const vector<unsigned int> &desiredMetrics) {

	bool ret = false;
	map<unsigned int, MetricPair>::const_iterator it =
			validCCVDimensions.find(dim);

	if (it != validCCVDimensions.end()) {
		MetricPair pair = it->second;
		for (size_t i=0; i < desiredMetrics.size(); i++) {
			if (pair.x == desiredMetrics[i] || pair.y == desiredMetrics[i]) {
				ret = true;
				break;
			}
		}
	} else {
		handle_error("[detection_engine]:  isCCVDimensionWanted(): "
						"dimension not found in CCV space");
	}

	return ret;
}

inline
double euclidean_dist(const vector<double> &x, const vector<double> &y)
{
	if (x.size() != y.size())
		handle_error("[detection_engine]:  in euclidean_dist(): "
				"vectors of different size");

	double ret=0;

	for (size_t i=0; i < x.size(); i++)
		ret = ret + pow((x[i] - y[i]),2.0);
	ret = sqrt(ret);

	if (ret < 0)
		handle_error("[detection_engine]:  in euclidean_dist(): "
				"distance cannot be negative");

	return ret;
}

double knn_ccv_distance(const vector <double> &ccv,
		const vector< vector<double> > &sphere)
{
	// set distance to the first one
	double dist = euclidean_dist(ccv, sphere[0]);

	for (size_t i=1; i < sphere.size(); i++) {
		double tmp = euclidean_dist(ccv, sphere[i]);
		if (tmp < dist)
			dist = tmp;
	}

	/*------------- Parallel code --------------*/
/*	double dist = euclidean_dist(ccv, sphere[0]);
	size_t i, data_size=sphere.size();
	double tmp, local_dist;
#pragma omp parallel default(shared) private(i,local_dist,tmp)
	{
	local_dist = dist;

#pragma omp for schedule(dynamic)
	for (i=1; i < data_size; i++) {
		tmp = euclidean_dist(ccv, sphere[i]);
		if (tmp < local_dist)
			local_dist = tmp;
	}

#pragma omp critical
	{
		if (local_dist < dist)
			dist = local_dist;
	}

	}*/

	return dist;
}

void calculate_knn_thresh()
{
	unsigned int i;
	unsigned int sample_size = 100;

	srand(time(NULL));
	vector<double> distances;

	for (i=0; i < sample_size; i++) {
		size_t h_sphere_size = h_sphere.size();
		size_t sample_index =
				(size_t)(((double)rand()/(double)RAND_MAX) *
						(double)h_sphere_size);

		//cout << "Index: " << sample_index << endl;

		vector<double> test_ccv = h_sphere[sample_index];

		// Eliminate testing point from hypher-sphere
		vector< vector<double> > new_sphere(h_sphere);
		new_sphere.erase(new_sphere.begin()+(sample_index));

		// Get distance
		double d = knn_ccv_distance(test_ccv, new_sphere);
		distances.push_back(d);
		cout << "Distance: " << d << endl;
	}

	// Statistics of KNN distances
	double mu = 0, max=0;
	for (size_t k=0; k < distances.size(); k++) {
		mu = mu + distances[k];
		if (distances[k] > max) max = distances[k];
	}
	mu = mu / static_cast<double>(distances.size());
	cout << "Trained KNN distances: AVG " << mu << " MAX " << max << endl;

	return;
}
